import os
from pathlib import Path, PureWindowsPath
import sys

import numpy as np
import pytest


@pytest.fixture
def unload_pandas():
    del globals()["pd"]
    for m in list(sys.modules.keys()):
        if m.startswith("pandas"):
            del sys.modules[m]


@pytest.fixture
def pd():
    import pandas as pd

    return pd


@pytest.fixture
def pandas_path():
    import pandas_path
    from pandas_path import path  # noqa

    assert pandas_path.__version__ is not None

    return pandas_path


@pytest.fixture
def sample_paths():
    return list(Path().glob("**/*.py"))


@pytest.fixture
def sample_series(pd, pandas_path, sample_paths):
    return pd.Series(str(s) for s in sample_paths)


def test_not_registered(unload_pandas, pd):
    assert not getattr(pd.Series, "path", None)


def test_registered(pd, pandas_path):
    assert getattr(pd.Series, "path", None)


def test_properties(sample_series, sample_paths):
    expected = [Path(f).name for f in sample_paths]
    assert sample_series.path.name.tolist() == expected

    expected = [str(Path(f).parent) for f in sample_paths]
    assert sample_series.path.parent.tolist() == expected

    expected = [str(Path(f).suffix) for f in sample_paths]
    assert sample_series.path.suffix.tolist() == expected

    expected = [str(Path(f).stem) for f in sample_paths]
    assert sample_series.path.stem.tolist() == expected


def test_methods(sample_series, sample_paths):
    # is_file
    expected = [Path(f).is_file() for f in sample_paths]
    assert sample_series.path.is_file().tolist() == expected

    # exists
    expected = [Path(f).exists() for f in sample_paths]

    # append a file that doesn't exist
    with_not_exist = sample_series.copy()
    with_not_exist.loc[-1] = "not_exist.file"
    expected += [False]

    assert with_not_exist.path.exists().tolist() == expected

    # all files except added one are .py
    assert with_not_exist.path.match("*.py").tolist() == expected

    # with_suffix
    expected = [str(Path(f).with_suffix(".png")) for f in sample_paths]
    assert sample_series.path.with_suffix(".png").tolist() == expected

    # with_name
    expected = [str(Path(f).with_name("new.png")) for f in sample_paths]
    assert sample_series.path.with_name("new.png").tolist() == expected


def test_operators(sample_series, sample_paths):
    # left hand
    expected = [str(Path(f).parent / "new_file.txt") for f in sample_paths]
    assert (sample_series.path.parent.path / "new_file.txt").tolist() == expected

    # right hand
    expected = [str("new_root" / Path(f)) for f in sample_paths]
    assert ("new_root" / sample_series.path).tolist() == expected

    # chaining - error
    with pytest.raises(TypeError):
        sample_series.path.parent.path / "new_folder" / "new_file"

    # chaining - explicit
    expected = [str(Path(f).parent / "new_sub_folder" / "new_file.txt") for f in sample_paths]
    assert (
        (sample_series.path.parent.path / "new_sub_folder").path / "new_file.txt"
    ).tolist() == expected

    # left-hand path object (only supported Python > 3.7 because of https://bugs.python.org/issue34775)
    if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
        assert (
            (Path("new_folder") / sample_series.path) == ("new_folder" + os.sep + sample_series)
        ).all()

    # right-hand path object
    assert (
        (sample_series.path.parent.path / Path("new_file.txt"))
        == (
            (sample_series.path.parent.replace(".", "") + os.sep).str.lstrip("." + os.sep)
            + "new_file.txt"
        )
    ).all()


def test_elementwise(pd, sample_series, sample_paths):
    # operators
    to_append = [f"file_{i}.txt" for i in range(len(sample_series))]
    to_prepend = [f"folder_{i}" for i in range(len(sample_series))]

    expected = [str(Path(f).parent / a) for f, a in zip(sample_paths, to_append)]
    assert (sample_series.path.parent.path / to_append).tolist() == expected
    assert (sample_series.path.parent.path / np.array(to_append)).tolist() == expected
    assert (sample_series.path.parent.path / pd.Series(to_append)).tolist() == expected
    assert (sample_series.path.parent.path / pd.Series(to_append).path).tolist() == expected

    expected = [str(p / Path(f)) for f, p in zip(sample_paths, to_prepend)]
    assert (to_prepend / sample_series.path).tolist() == expected
    assert (pd.Series(to_prepend).path / sample_series.path).tolist() == expected

    # These will not work :/
    # assert (np.array(to_prepend) / sample_series.path).tolist() == expected
    # assert (pd.Series(to_prepend) / sample_series.path).tolist() == expected

    # methods
    # with_suffix
    suffixes = np.random.choice([".txt", ".py", ".jpeg"], size=len(sample_series))

    expected = [str(Path(f).with_suffix(s)) for f, s in zip(sample_paths, suffixes)]
    assert sample_series.path.with_suffix(suffixes).tolist() == expected


def test_custom_accessor(pd, sample_paths):
    from pandas_path import register_path_accessor

    register_path_accessor("win", PureWindowsPath)

    win_series = pd.Series(
        [
            "c:\\test\\f1.txt",
            "c:\\test2\\f2.txt",
        ]
    )

    assert (win_series.win.parent == (win_series.str.rsplit("\\", 1).str[0])).all()

    assert win_series.win.__name__ == "PureWindowsPathAccessor"
    assert "PureWindowsPath" in win_series.win.__doc__



from types import FunctionType, LambdaType, MethodType

import numpy as np
import pandas as pd


def path_accessor_factory(path_class, *args, **kwargs):
    class PathAccessor:
        _doc_str_f = """A `.{accessor_name}` accessor on Series and Index objects so that all of the methods
            from {path_class} are available.

            Note: Always will return a `str` rather than a {path_class} object so that joining with existing
                dataframes, etc. will continue to work.

            Note: May not be performant on extremely long lists since we cast to `{path_class}`
            for every operation.

        Raises
        ------
        AttributeError
            If the method we try to use on the accessor is not a function or property or
            is not an attribute on {path_class}, raise an AttributeError.
        """

        def __init__(self, pandas_obj):
            self._validate(pandas_obj)
            self._obj = pandas_obj

        @staticmethod
        def _validate(obj):
            [path_class(x, *args, **kwargs) for x in obj.values]

        @staticmethod
        def _to_path_object(path_str):
            return path_class(path_str, *args, **kwargs)

        def __getattr__(self, attr):
            # return accessor name for clarity
            if attr == "__name__":
                return self.__class__.__name__

            apply_series = self._obj.to_series() if isinstance(self._obj, pd.Index) else self._obj

            # check the type of this attribute on a Path object (we need an actual instance) since
            # the super classes dispatch
            if isinstance(self._obj.values[0], path_class):
                attr_type = getattr(type(self._obj.values[0]), attr, None)
            else:
                attr_type = getattr(type(self._to_path_object(self._obj.values[0])), attr, None)

            # if we're asking for a property, do the calculation and return the result
            if isinstance(attr_type, property):

                def _to_apply(x):
                    res = getattr(self._to_path_object(x), attr)
                    return res if not isinstance(res, path_class) else str(res)

                return apply_series.apply(_to_apply)

            # if we're asking for a function, return a callable method
            elif isinstance(attr_type, (FunctionType, MethodType, LambdaType)):

                def _callable(*args, **kwargs):
                    if len(args) == 1 and isinstance(
                        args[0], (list, tuple, np.ndarray, pd.Series, pd.Index)
                    ):
                        other = args[0]
                        return self._elementwise(other, attr)

                    else:

                        def _to_apply(x):
                            res = getattr(self._to_path_object(x), attr)(*args, **kwargs)
                            return res if not isinstance(res, path_class) else str(res)

                        return apply_series.apply(_to_apply)

                return _callable

            else:
                raise AttributeError(f"Can't find or handle path attribute {attr}.")

        def _elementwise(self, other, attr):
            if isinstance(other, PathAccessor):
                other = other._obj

            other_array = np.array(other)

            if other_array.ndim > 1:
                raise ValueError(
                    "Can only do an elementwise operations with a 1d array, list, or Series."
                )

            if other_array.shape[0] != len(self._obj):
                raise ValueError(
                    "Can only do an elementwise operation with arrays of the same length."
                )

            def _to_apply(this_elem, other_elem):
                res = getattr(self._to_path_object(this_elem), attr)(other_elem)
                return res if not isinstance(res, path_class) else str(res)

            elementwise_result = pd.Series(
                [
                    _to_apply(this_elem, other_elem)
                    for this_elem, other_elem in zip(self._obj.values, other_array)
                ],
                index=self._obj.index,
            )

            return elementwise_result

        def _path_join(self, other, side):
            if isinstance(other, (list, tuple, np.ndarray, pd.Series, pd.Index, PathAccessor)):
                return self._elementwise(other, side)
            else:
                return self.__getattr__(side)(other)

        # operators don't override correctly unless defined
        def __truediv__(self, other):
            return self._path_join(other, "__truediv__")

        def __rtruediv__(self, other):
            return self._path_join(other, "__rtruediv__")

    return PathAccessor


def register_path_accessor(accessor_name, path_class, *args, **kwargs):
    accessor_class = path_accessor_factory(path_class, *args, **kwargs)

    # set name and docstring dynamically for reference
    accessor_class.__name__ = path_class.__name__ + "Accessor"
    accessor_class.__doc__ = accessor_class._doc_str_f.format(
        path_class=path_class.__name__, accessor_name=accessor_name
    )

    pd.api.extensions.register_series_accessor(accessor_name)(accessor_class)
    pd.api.extensions.register_index_accessor(accessor_name)(accessor_class)



#!/usr/bin/env python
from pathlib import Path

import setuptools
from setuptools import setup


def load_reqs(path):
    reqs = []
    with open(path, "r") as f:
        for line in f.readlines():
            if line.startswith("-r"):
                reqs += load_reqs(line.split(" ")[1].strip())
            else:
                req = line.strip()
                if req and not req.startswith("#"):
                    reqs.append(req)
    return reqs


req_path = Path(__file__).parent / "requirements.txt"
requirements = load_reqs(req_path)

long_description = open(Path(__file__).parent / "README.md").read()


setup(
    name="pandas_path",
    url="https://github.com/drivendataorg/pandas-path",
    project_urls={
        "Source Code": "https://github.com/drivendataorg/pandas-path",
        "DrivenData": "http://drivendata.co",
    },
    version="0.3.0",
    author="DrivenData",
    author_email="info@drivendata.org",
    include_package_data=True,
    description="Pathlib functionality for pandas.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=setuptools.find_packages(),
    keywords=["data science"],
    python_requires=">=3.6",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
